// This is needed for opening files bigger than 2GB
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

/* from binfsrch.c by CBF */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>


/* The difference between a binary and a text file, on read,
is the conversion of end-of-line delimiters. What those
delimiters are does not affect the action. In some cases
the presence of 0x1a EOF markers (MsDos) does.

This is a version of Knuth-Morris-Pratt algorithm. The
point of using this is to avoid any backtracking in file
reading, and thus avoiding any use of buffer arrays.
*/

u_int64_t chrcount=0; /* count of input chars, zeroed */
char marker[31];
int *next=NULL;
int lgh;
int ch;
int items; /* count of markers found */
FILE *ffl=NULL;
long loc_start_addr=0;

/* --------------------- */

/* Almost straight out of Sedgewick */
/* The next array indicates what index in id should next be
compared to the current char. Once the (lgh - 1)th char
has been successfully compared, the id has been found.
The array is formed by comparing id to itself. */

void initnext(int *next, const char *id, int lgh)
{
	int i, j;
	
	assert(lgh > 0);
	next[0] = -1; i = 0; j = -1;
	while (i < lgh) {
		while ((j >= 0) && (id[i] != id[j])) j = next[j];
		i++; j++;
		next[i] = j;
	}
//	#ifdef DEBUG
//	for (i = 0; i <= lgh; i++)
//		printf("id[%d] = '%c' next[%d] = %d\n", i, id[i], i, next[i]);
//	#endif
} /* initnext */

/* --------------------- */

/* reads f without rewinding until either EOF or *marker
has been found. Returns EOF if not found. At exit the
last matching char has been read, and no further. */
int kmpffind(const char *marker, int lgh, int *next, FILE *fl)
{
	int j; /* char position in marker to check */
	int ch; /* current char */

	assert(lgh > 0);
	j = 0;
	while ((j < lgh) && (EOF != (ch = getc(fl)))) {
		chrcount++;
//		printf("j: %d, ch: %c, marker[%d] = %c, next %d\n",j,ch,j,marker[j],next[j]);//DBG
		while ((j >= 0) && (ch != marker[j])) j = next[j];
		j++;
	}
	return ch;
} /* kmpffind */

/* --------------------- */

/* Function that initialize search state.
 * (reopen file and null counter)
 * if start_addr is not null, function seeks start 
 * searching from start_addr. This is used as optimization when
 * looking for modules, because modules are usually beyond 0x30000000
*/

int search_init(char *inputfile2/*,FILE *f2*/,const char *marker2,u_int64_t start_addr) {

//	printf("search_init: %s, %s\n",inputfile2,marker); // DBG
	memcpy(marker,marker2,strlen(marker2)); 
	chrcount=0;

	if (!(ffl = fopen64(inputfile2, "rb"))) {
		// _FILE_OFFSET_BITS should be enough,
		// but seems that explicit fopen64 is required
		fprintf(stderr, "Can't open %s, %s\n", inputfile2, strerror(errno));
		return -1;
	}

	if (start_addr != 0) { //jump to offset
//		printf("!!! Seeking 0x%lx !!!\n",start_addr); //DBG
		loc_start_addr=start_addr;
		fseek(ffl,start_addr,SEEK_SET); 
	}	
//	printf("!!# pos: %lx #!!\n",ftell(ffl));

	lgh = strlen(marker);
	if (!(next = malloc((1 + lgh) * sizeof *next))) {// +1 -> \0 ??
		puts("No memory");
		return -2;
	}	
//	printf("search_init: lgh:%d\n", lgh);	//DBG
	
	initnext(next, marker, lgh);

	
	items = 0;
	return 0;
}

/* Function that finds next occurence of string
 * returns position
 */

u_int64_t search_next()
{	
//	printf("search_next: %d %s : %x\n", items, marker,chrcount-strlen(marker)); //DBG
	if (EOF == kmpffind(marker, lgh, next, ffl)) { //find next
		free(next);
		next=NULL;
		return 0;
	}
	return chrcount-strlen(marker)+loc_start_addr;

} 

/* Function for free & close 
 *
 */

int search_exit()
{	
	if (ffl != NULL) {fclose(ffl); ffl=NULL;};
	if (next != NULL) {free(next); next=NULL;}
	return 0;

}
/* binfsrch */




